WITH filtered_users AS (
    SELECT * FROM user_management_service_user
    WHERE username = 'Username1'
),
filtered_savestates AS (
    SELECT ss.* 
    FROM user_management_service_savestate ss
    INNER JOIN filtered_users uu ON ss.user_id = uu.id
),
filtered_querystates AS (
    SELECT qs.* 
    FROM user_management_service_querystate qs
    INNER JOIN filtered_savestates ss ON ss.id = qs.save_state_id
),
raw_nodes AS (
    SELECT node
    FROM (
        SELECT jsonb_build_object(
            '_id', concat('user_management_service_savestate-', id::text),
            'label', 'user_management_service_savestate',
            'attributes', to_jsonb(filtered_savestates.*) - 'id'
        ) AS node
        FROM filtered_savestates

        UNION ALL

        SELECT jsonb_build_object(
            '_id', concat('user_management_service_user-', id::text),
            'label', 'user_management_service_user',
            'attributes', to_jsonb(filtered_users.*) - 'id'
        ) AS node
        FROM filtered_users

        UNION ALL

        SELECT jsonb_build_object(
            '_id', concat('user_management_service_querystate-', id::text),
            'label', 'user_management_service_querystate',
            'attributes', to_jsonb(filtered_querystates.*) - 'id'
        ) AS node
        FROM filtered_querystates
    ) sub
    LIMIT 5000
),

nodes_cte AS (
    SELECT jsonb_agg(node) AS nodes
    FROM raw_nodes
),

edges_cte AS (
    SELECT jsonb_agg(edge) AS edges
    FROM (
        SELECT jsonb_agg(
            jsonb_build_object(
                '_id', concat('user_management_service_savestate-', ss.id::text, '-user_management_service_user-', uu.id::text),
                'label', 'user_id.id',
                'attributes', '{}'::jsonb,
                'from', ss.id::text,
                'to', uu.id::text
            )
        ) AS edge
        FROM filtered_savestates ss
        INNER JOIN filtered_users uu ON ss.user_id = uu.id

        UNION ALL

        SELECT jsonb_agg(
            jsonb_build_object(
                '_id', concat('user_management_service_savestate-', ss.id::text, '-user_management_service_querystate-', qs.id::text),
                'label', 'id.save_state_id',
                'attributes', '{}'::jsonb,
                'from', ss.id::text,
                'to', qs.id::text
            )
        ) AS edge
        FROM filtered_savestates ss
        INNER JOIN filtered_querystates qs ON ss.id = qs.save_state_id
    ) aggregated_edges
),

node_counts AS (
    SELECT label, COUNT(*) AS count
    FROM (
        SELECT 'user_management_service_savestate' AS label FROM filtered_savestates
        UNION ALL
        SELECT 'user_management_service_user' AS label FROM filtered_users
        UNION ALL
        SELECT 'user_management_service_querystate' AS label FROM filtered_querystates
    ) t
    GROUP BY label
),

edge_counts AS (
    SELECT 'user_id.id' AS label, COUNT(*) AS count
    FROM filtered_savestates
    GROUP BY label

    UNION ALL

    SELECT 'id.save_state_id' AS label, COUNT(*) AS count
    FROM filtered_querystates
    GROUP BY label
)

SELECT jsonb_build_object(
    'nodes', nodes_cte.nodes,
    'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
    'nodeCounts', COALESCE(
        (SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count)) FROM node_counts),
        '[]'::jsonb
    ),
    'edgeCounts', COALESCE(
        (SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count)) FROM edge_counts),
        '[]'::jsonb
    )
) AS result
FROM nodes_cte, edges_cte;