import type { AllLogicStatement, AnyStatement, BackendQueryFormat, NodeStruct, QueryStruct } from 'ts-common';
import type { QueryText } from '../model';
import { extractLogicSQL } from './convertLogicToSQL';
import type { QueryCacheData } from '../cypher/converter/model';

function cleanStringOfNewLinesAndTabs(str: string): string {
  if (!str) {
    return '';
  }
  let trimmedSQL = str.replace(/\n/g, ' ');
  trimmedSQL = trimmedSQL.replaceAll(/ {2,50}/g, ' ');
  trimmedSQL = trimmedSQL.replace(/\t+/g, '');
  return trimmedSQL.trim();
}

// function buildNodeCTE(path: QueryStruct, logic: AnyStatement | undefined): { node: string; edges: string; rows: string[] } {
//   const node = path.node;
//   const pathId = path.id;

//   if (!node || !node.label || !node.id) {
//     return { node: '', edges: '', rows: [] };
//   }

//   const loopThoughNodes = (node: NodeStruct): NodeStruct[] => {
//     if (node.relation && node.relation.node) {
//       return [node, ...loopThoughNodes(node.relation.node)];
//     }
//     return [node];
//   };
//   const allNodes = loopThoughNodes(node);

//   const where = logic ? extractLogicSQL(logic, { nodes: allNodes }).logic : undefined;
//   const rowList = allNodes.map(n => `${n.id}_row`);
//   const select = `
//   SELECT ${[
//     ...allNodes.map(n => `to_jsonb(${n.id}) AS ${n.id}_row`),
//     ...allNodes
//       .filter(n => n.relation && n!.relation!.node)
//       .map(
//         n =>
//           `${n.id}.${n!.relation!.label!.split('.')[0]}::text AS ${n.id}_${n!.relation!.label!.split('.')[0]}, ${n!.relation!.node!.id}.${
//             n!.relation!.label!.split('.')[1]
//           }::text AS ${n!.relation!.node!.id}_${n!.relation!.label!.split('.')[1]}`,
//       ),
//   ].join(', ')}
//   FROM ${node.label} ${node.id}
//   ${allNodes
//     .filter(n => n?.relation?.node?.label && n?.relation?.node?.id)
//     .map(
//       n =>
//         `INNER JOIN ${n!.relation!.node!.label} ${n!.relation!.node!.id} ON ${n.id}.${n!.relation!.label!.split('.')[0]}::text = ${
//           n!.relation!.node!.id
//         }.${n!.relation!.label!.split('.')[1]}::text`,
//     )
//     .join('\n')}
//     ${where ? `WHERE ${where}` : ''}`;

//   const edges = buildEdgesCTE(node, rowList, pathId).join(' UNION ALL ');

//   return { node: select, edges, rows: rowList };
// }

// function buildEdgesCTE(node: NodeStruct | undefined, rowList: string[], pathId: string): string[] {
//   if (!node || !node.relation || !node.relation.node || !node.id || !node.relation.node.id || !node.relation.label) {
//     return [];
//   }

//   const edge = `
//   SELECT jsonb_build_object(
//     '_id', concat('${node.label}-', ${node.id}_${node.relation.label.split('.')[0]}::text, '-${node.relation.node.label}-', ${
//     node.relation.node.id
//   }_${node.relation.label.split('.')[1]}::text),
//     'label', '${node.relation.label}',
//     'attributes', '{}'::jsonb,
//     'from', ${node.id}_${node.relation.label.split('.')[0]}::text,
//     'to', ${node.relation.node.id}_${node.relation.label.split('.')[1]}::text
//   ) AS edge,
//   ${rowList.join(',\n')}
//   FROM ${pathId}`;

//   const ret = cleanStringOfNewLinesAndTabs(edge);
//   return [ret, ...buildEdgesCTE(node.relation.node, rowList, pathId)];
// }

// function buildNodeCountsCTE(node: NodeStruct | undefined): string[] {
//   if (!node || !node.label || !node.id) {
//     return [];
//   }
//   const nodeParts: string[] = [];
//   const mainNode = `SELECT '${node.label}' AS label FROM ${node.label}`;
//   nodeParts.push(cleanStringOfNewLinesAndTabs(mainNode));
//   buildNodeCountsCTE(node.relation?.node).forEach(part => nodeParts.push(part));
//   return nodeParts;
// }

// function buildEdgeCountsCTE(node: NodeStruct | undefined): string[] {
//   if (!node || !node.relation || !node.relation.node || !node.id || !node.relation.node.id || !node.relation.label) {
//     return [];
//   }

//   const edge = `
//   SELECT '${node.relation.label}' AS label, COUNT(*) AS count
//   FROM ${node.label} ${node.id}
//   JOIN ${node.relation.node.label} ${node.relation.node.id}
//   ON ${node.id}.${node.relation.label.split('.')[0]} = ${node.relation.node.id}.${node.relation.label.split('.')[1]}
//   GROUP BY label`;
//   const ret = cleanStringOfNewLinesAndTabs(edge);
//   return [ret, ...buildEdgeCountsCTE(node.relation.node)];
// }

export function query2SQL(query: BackendQueryFormat): QueryText {
  const loopThoughNodes = (node: NodeStruct): NodeStruct[] => {
    if (node.relation && node.relation.node) {
      return [node, ...loopThoughNodes(node.relation.node)];
    }
    return [node];
  };
  const allNodes = query.query.flatMap(path => loopThoughNodes(path.node));
  const allUniqueNodes: Record<
    string,
    { label: string; joins: { from: string; leftJoin: string; rightJoin: string; to: string; fromLabel: string; toLabel: string }[] }
  > = Object.fromEntries(
    allNodes
      .map(n => [n.id, { label: n.label, joins: [] }])
      .filter((node, index, self) => {
        return index === self.findIndex(t => t[0] === node[0] && t[1] === node[1]);
      }),
  );
  allNodes.forEach(n => {
    if (n.id && n.relation && n.relation.node && n.relation.node.id && n.label && n.relation.node.label) {
      allUniqueNodes[n.id].joins.push({
        from: n.id,
        leftJoin: n!.relation!.label!.split('.')[0],
        rightJoin: n!.relation!.label!.split('.')[1],
        to: n!.relation!.node!.id,
        fromLabel: n!.label,
        toLabel: n!.relation!.node!.label,
      });
    }
  });

  console.log(JSON.stringify(allNodes, null, 2));

  const nodeEntries = Object.entries(allUniqueNodes);
  const where = query.logic ? extractLogicSQL(query.logic, { nodes: allNodes }).logic : undefined;

  const firstNode = nodeEntries[0][1];

  const relIDSql = Object.values(allUniqueNodes)
    .map(n =>
      n.joins.map(j => `${j.from}.${j.leftJoin} AS ${j.from}_${j.leftJoin}, ${j.to}.${j.rightJoin} AS ${j.to}_${j.rightJoin}`).join(', '),
    )
    .filter(s => s.length > 0);
  const select = [...nodeEntries.map(n => `to_jsonb(${n[0]}) AS ${n[0]}_row`), ...relIDSql].join(', ');

  let sql = `SELECT ${select} FROM ${firstNode.label} ${nodeEntries[0][0]}`;
  // let sql = `SELECT jsonb_agg(row_to_json(t)) FROM ${firstNode.label} ${nodeEntries[0][0]}`;

  let joins = '';
  for (let i = 0; i < nodeEntries.length; i++) {
    const [nodeId, node] = nodeEntries[i];
    if (node.joins) {
      node.joins.forEach(j => {
        joins += ` INNER JOIN ${allUniqueNodes[j.to].label} ${j.to} ON ${j.from}.${j.leftJoin}::text = ${j.to}.${j.rightJoin}::text`;
      });
    }
  }

  sql += joins;
  sql += where ? ` WHERE ${where}` : '';
  const sqlNoLimit = sql;
  sql += query.limit ? ` LIMIT ${query.limit}` : '';

  const nodes_cte = `SELECT jsonb_agg(row_to_json(t)) AS nodes FROM (${nodeEntries
    .map(n => `SELECT DISTINCT ${n[0]}_row AS node FROM combined`)
    .join(' UNION ALL ')}) AS t`;

  const allUniqueEdges = Object.values(allUniqueNodes)
    .map(n =>
      n.joins
        .map(j => {
          return `SELECT jsonb_build_object(
          '_id', concat('${n.label}-', md5(${j.from}_row::text), '-${allUniqueNodes[j.to].label}-', md5(${j.to}_row::text)),
          'label', '${j.leftJoin}.${j.rightJoin}',
          'attributes', '{}'::jsonb,
          'from', md5(${j.from}_row::text),
          'to', md5(${j.to}_row::text)
        ) AS edge
        FROM combined`;
        })
        .filter(s => s.length > 0)
        .join(' UNION ALL '),
    )
    .filter(s => s.length > 0)
    .join(' UNION ALL ');

  // console.log(JSON.stringify(allUniqueNodes, null, 2));
  // const rowList = allNodes.map(n => `${n.id}_row`);
  // const select = `
  // SELECT ${allNodes.map(n => `to_jsonb(${n.id}) AS ${n.id}_row`).join(', ')},
  // ${allNodes
  //   .filter(n => n.relation && n!.relation!.node)
  //   .map(
  //     n =>
  //       `${n.id}.${n!.relation!.label!.split('.')[0]}::text AS ${n.id}_${n!.relation!.label!.split('.')[0]}, ${n!.relation!.node!.id}.${
  //         n!.relation!.label!.split('.')[1]
  //       }::text AS ${n!.relation!.node!.id}_${n!.relation!.label!.split('.')[1]}`,
  //   )
  //   .join(', ')}
  // FROM ${node.label} ${node.id}
  // ${allNodes
  //   .filter(n => n?.relation?.node?.label && n?.relation?.node?.id)
  //   .map(
  //     n =>
  //       `INNER JOIN ${n!.relation!.node!.label} ${n!.relation!.node!.id} ON ${n.id}.${n!.relation!.label!.split('.')[0]}::text = ${
  //         n!.relation!.node!.id
  //       }.${n!.relation!.label!.split('.')[1]}::text`,
  //   )
  //   .join('\n')}
  //   ${where ? `WHERE ${where}` : ''}`;

  // const edges = buildEdgesCTE(node, rowList, pathId).join(' UNION ALL ');

  // const nodesCTE = query.query.flatMap(path => ({ ...buildNodeCTE(path, query.logic), path: path }));
  // const nodeCountsCTE = [...new Set(query.query.flatMap(path => buildNodeCountsCTE(path.node)))];
  // const edgeCountsCTE = [...new Set(query.query.flatMap(path => buildEdgeCountsCTE(path.node)))];

  const mainQuery = `WITH combined AS (${sql}),
    nodes_cte AS (
      SELECT jsonb_agg(row_to_json(t)) AS nodes FROM (${nodeEntries
        .map(
          ([nodeId, node]) => `
          SELECT 
            md5(${nodeId}_row::text) AS _id,
            '${node.label}' AS label,
            ${nodeId}_row - 'id' AS attributes
          FROM combined`,
        )
        .join(' UNION ALL ')}) AS t
    ),
    edges_cte AS (
      SELECT jsonb_agg(edge) AS edges
        FROM (
          ${allUniqueEdges.length > 0 ? allUniqueEdges : 'SELECT null AS edge WHERE FALSE'}
        ) AS e
    ),
    node_counts AS (
      SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count)) AS nodeCounts
      FROM (
        SELECT label, COUNT(*) AS count
        FROM (
          ${[...new Set(nodeEntries.map(([_, node]) => node.label))]
            .map(label => `SELECT '${label}' AS label FROM ${label}`)
            .join(' UNION ALL ')}
        ) AS node_labels
        GROUP BY label
      ) AS counts
    )
    SELECT jsonb_build_object(
      'nodes', nodes_cte.nodes,
      'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
      'nodeCounts', COALESCE((SELECT nodeCounts FROM node_counts), '[]'::jsonb)
    ) AS result FROM nodes_cte, edges_cte;`;

  // 'nodeCounts', COALESCE( (SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
  //   FROM node_counts), '[]'::jsonb ),
  // 'edgeCounts', COALESCE( (SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
  //   FROM edge_counts), '[]'::jsonb )
  // node_counts AS (
  //   SELECT label, COUNT(*) AS count
  //   FROM (
  //     ${nodesCTE.map(node => node.node).join(' UNION ALL ')}
  //   ) t
  //   GROUP BY label
  // ),
  // edge_counts AS (
  //   ${nodesCTE.map(node => node.edges).join(' UNION ALL ')}
  // )

  return {
    query: cleanStringOfNewLinesAndTabs(mainQuery),
    countQuery: '',
  };
}
